#include "wapp.h"
#include "common.h"
#include "raylib.h"
#include <stdint.h>
#include <stdbool.h>
#include <assert.h>
#include <string.h>

#define MOVABLE_TAG_SHIFT     0
#define RENDERABLE_TAG_SHIFT  1
#define COLLIDER_TAG_SHIFT    2
#define INSIDE_ZONE_TAG_SHIFT 3

enum EntityTag {
  ENTITY_TAG_MOVABLE     = 1 << MOVABLE_TAG_SHIFT,
  ENTITY_TAG_RENDERABLE  = 1 << RENDERABLE_TAG_SHIFT,
  ENTITY_TAG_COLLIDER    = 1 << COLLIDER_TAG_SHIFT,
  ENTITY_TAG_INSIDE_ZONE = 1 << INSIDE_ZONE_TAG_SHIFT,
};

typedef struct Entity Entity;
struct Entity {
  u16 id;
};

typedef struct Position Position;
struct Position {
  f32 x;
  f32 y;
};

typedef struct Scale Scale;
struct Scale {
  f32 width;
  f32 height;
};

typedef struct Rect Rect;
struct Rect {
  Position position;
  Scale scale;
};

typedef struct Velocity Velocity;
struct Velocity {
  f32 x;
  f32 y;
};

typedef struct Manager Manager;
struct Manager {
  Entity   *entities;
  Rect     *rects;
  Velocity *velocities;
  u8       *tags;
  u64      count;
};

typedef void (*ScaleInitialiser)(Scale *scale, XOR256State *state);
typedef void (*VelocityInitialiser)(Velocity *velocity, XOR256State *state);
typedef void (*RaylibDrawRectFunc)(int posX, int posY, int width, int height, Color color);

void init_manager(const Allocator *allocator, Manager *manager);
void init_position(Position *position, XOR256State *state);
void init_scale_wanderer(Scale *scale, XOR256State *state);
void init_scale_zone(Scale *scale, XOR256State *state);
void init_velocity(Velocity *velocity, XOR256State *state);
void zero_velocity(Velocity *velocity, XOR256State *state);
void update_positions(u8 *tags, Rect *rects, Velocity *velocities, u64 count);
void render_entities(const u8 *tags, const Rect *rects, u64 count);
u64 collides(const Rect *rect, const Rect *collider);
f32 get_random_float(XOR256State *state);

int main(void) {
  SetTraceLogLevel(LOG_NONE);
  InitWindow(WIDTH, HEIGHT, "DOD test");
  // SetTargetFPS(120);

  Allocator arena = wapp_mem_arena_allocator_init(MB(20));
  assert(!wapp_mem_allocator_invalid(&arena));

  XOR256State state = wapp_prng_xorshift_init_state();
  Manager manager   = {0};

  init_manager(&arena, &manager);

  ScaleInitialiser scale_initialisers[2]       = {init_scale_wanderer, init_scale_zone};
  VelocityInitialiser velocity_initialisers[2] = {init_velocity, zero_velocity};

  for (u64 i = 0; i < manager.count; ++i) {
    u8 is_zone = (i - ZONE_COUNT) >> 63;

    manager.entities[i].id  = i;
    manager.tags[i]         = (ENTITY_TAG_COLLIDER & (is_zone << COLLIDER_TAG_SHIFT)) | ENTITY_TAG_MOVABLE | ENTITY_TAG_RENDERABLE;

    init_position(&(manager.rects[i].position), &state);
    scale_initialisers[is_zone](&(manager.rects[i].scale), &state);
    velocity_initialisers[is_zone](&(manager.velocities[i]), &state);
  }

  while (!WindowShouldClose()) {
    f64 time = GetTime();
    if (time >= 20.0) {
      break;
    }

    update_positions(manager.tags, manager.rects, manager.velocities, manager.count);

    BeginDrawing();

    ClearBackground(BG_COLOR);

    // render_entities(manager.tags, manager.rects, manager.count);

    DrawFPS(10, 10);

    EndDrawing();
  }

  wapp_mem_arena_allocator_destroy(&arena);

  CloseWindow();

  return 0;
}

void init_manager(const Allocator *allocator, Manager *manager) {
  assert(allocator != NULL && manager != NULL);

  u64 total_count      = (u64)WANDERER_COUNT + (u64)ZONE_COUNT;
  u64 entities_size    = sizeof(Entity)   * total_count;
  u64 rects_size       = sizeof(Rect)     * total_count;
  u64 velocities_size  = sizeof(Velocity) * total_count;
  u64 tags_size        = sizeof(u8)       * total_count;
  u64 allocation_size  = entities_size + rects_size + velocities_size + tags_size;
  u8 *buffer           = wapp_mem_allocator_alloc(allocator, allocation_size);

  assert(buffer != NULL);

  memset(buffer, 0, allocation_size);

  manager->entities   = (Entity *)buffer;
  manager->rects      = (Rect *)(buffer + entities_size);
  manager->velocities = (Velocity *)(buffer + entities_size + rects_size);
  manager->tags       = (u8 *)(buffer + entities_size + rects_size + velocities_size);
  manager->count      = total_count;
}

void init_position(Position *position, XOR256State *state) {
  position->x = wapp_prng_xorshift_256(state) % WIDTH;
  position->y = wapp_prng_xorshift_256(state) % HEIGHT;
}

void init_scale_wanderer(Scale *scale, XOR256State *state) {
  f32 value = (f32)((wapp_prng_xorshift_256(state) % (MAX_WANDERER_DIM + 1 - MIN_WANDERER_DIM)) + MIN_WANDERER_DIM);
  scale->width  = value;
  scale->height = value;
}

void init_scale_zone(Scale *scale, XOR256State *state) {
  scale->width  = wapp_prng_xorshift_256(state) % ((u64)HALF_WIDTH - MIN_ZONE_DIM) + MIN_ZONE_DIM;
  scale->height = wapp_prng_xorshift_256(state) % ((u64)HALF_HEIGHT - MIN_ZONE_DIM) + MIN_ZONE_DIM;
}

void init_velocity(Velocity *velocity, XOR256State *state) {
  velocity->x = get_random_float(state);
  velocity->y = get_random_float(state);
}

void zero_velocity(Velocity *velocity, XOR256State *state) {
  velocity->x = 0.0f;
  velocity->y = 0.0f;
}

void update_positions(u8 *tags, Rect *rects, Velocity *velocities, u64 count) {
  u8 index = 0;
  f32 multipliers[2] = {1.0f, 0.5f};
  persistent u64 inside_zone_mask = 0x7;

  for (u64 i = ZONE_COUNT; i < count; ++i) {
    tags[i] &= inside_zone_mask;
    for (u64 j = 0; j < ZONE_COUNT; ++j) {
      tags[i] |= collides(&rects[i], &rects[j]) << INSIDE_ZONE_TAG_SHIFT;
    }

    index = (tags[i] & ENTITY_TAG_INSIDE_ZONE) >> INSIDE_ZONE_TAG_SHIFT;

    rects[i].position.x += velocities[i].x * multipliers[index];
    rects[i].position.y += velocities[i].y * multipliers[index];

    f32 max_x = WIDTH - rects[i].scale.width;
    f32 max_y = HEIGHT - rects[i].scale.height;

    if (rects[i].position.x < 0 || rects[i].position.x >= max_x) {
      rects[i].position.x = min(max(rects[i].position.x, 0), max_x);
      velocities[i].x *= -1;
    }

    if (rects[i].position.y < 0 || rects[i].position.y >= max_y) {
      rects[i].position.y = min(max(rects[i].position.y, 0), max_y);
      velocities[i].y *= -1;
    }
  }
}

void render_entities(const u8 *tags, const Rect *rects, u64 count) {
  persistent RaylibDrawRectFunc renderers[2] = {DrawRectangle, DrawRectangleLines};
  persistent Color colors[2]                 = {FG_COLOR, ZONE_COLOR};

  for (u64 i = 0; i < count; ++i) {
    u8 func_index  = (tags[i] & ENTITY_TAG_COLLIDER) >> COLLIDER_TAG_SHIFT;
    u8 color_index = func_index | ((tags[i] & ENTITY_TAG_INSIDE_ZONE) >> INSIDE_ZONE_TAG_SHIFT);

    renderers[func_index](
      rects[i].position.x,
      rects[i].position.y,
      rects[i].scale.width,
      rects[i].scale.height,
      colors[color_index]
    );
  }
}

u64 collides(const Rect *rect, const Rect *collider) {
  f32 rect_min_x = rect->position.x + rect->scale.width;
  f32 rect_min_y = rect->position.y + rect->scale.height;
  f32 rect_max_x = rect->position.x;
  f32 rect_max_y = rect->position.y;

  f32 zone_x0 = collider->position.x;
  f32 zone_y0 = collider->position.y;
  f32 zone_x1 = collider->position.x + collider->scale.width;
  f32 zone_y1 = collider->position.y + collider->scale.height;

  return (u64)(rect_min_x > zone_x0 && rect_min_y > zone_y0 &&
          rect_max_x < zone_x1 && rect_max_y < zone_y1);
}

f32 get_random_float(XOR256State *state) {
  i64 random = (i64)wapp_prng_xorshift_256(state) - INT64_MAX;
  return (f32)(random) / (f32)(INT64_MAX);
}
